import copy
from dataclasses import dataclass
import re
from typing import Any, Dict, List, Optional, Tuple, Union
import uuid

from common import (
    MULTI_LINE_REQUIREMENTS,
    OPENAPI_NO_VALIDATION,
    RequirementsFile,
    SINGLE_LINE_REQUIREMENTS,
    TEST_COMPUTE_CONFIG_DICT,
    TEST_CONTAINERFILE,
    TEST_REQUIREMENTS_FILES,
    TestLogger,
)
import pytest

from anyscale import Anyscale
from anyscale._private.anyscale_client import (
    FakeAnyscaleClient,
    WORKSPACE_CLUSTER_NAME_PREFIX,
)
from anyscale._private.models.image_uri import ImageURI
from anyscale._private.sdk.timer import FakeTimer
from anyscale.client.openapi_client.models import (
    Cloud,
    ComputeNodeType,
    ComputeTemplate,
    ComputeTemplateConfig,
)
from anyscale.compute_config import ComputeConfig, HeadNodeConfig
from anyscale.sdk.anyscale_client.models import (
    AccessConfig,
    ApplyServiceModel,
    GrpcProtocolConfig,
    ProductionServiceV2VersionModel,
    Protocols,
    RayGCSExternalStorageConfig as ExternalAPIRayGCSExternalStorageConfig,
    ServiceConfig as ExternalAPIServiceConfig,
    ServiceEventCurrentState,
    ServiceModel,
    ServiceVersionState as ExternalAPIServiceVersionState,
    TracingConfig as ExternalAPITracingConfig,
)
from anyscale.sdk.anyscale_client.models.cluster_environment_build import (
    ClusterEnvironmentBuild,
)
from anyscale.sdk.anyscale_client.models.cluster_environment_build_status import (
    ClusterEnvironmentBuildStatus,
)
from anyscale.service import ServiceSDK
from anyscale.service.models import (
    RayGCSExternalStorageConfig,
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionState,
    ServiceVersionStatus,
    TracingConfig,
)
from anyscale.utils.workspace_notification import (
    WorkspaceNotification,
    WorkspaceNotificationAction,
)


@pytest.fixture()
def service_sdk_with_fakes(
    sdk_with_fakes: Tuple[Anyscale, FakeAnyscaleClient, TestLogger, FakeTimer]
) -> Tuple[ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer]:
    sdk, client, logger, timer = sdk_with_fakes
    return sdk.service, client, logger, timer


class TestDeploy:
    def test_validation(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )

        with pytest.raises(TypeError):
            sdk.deploy(config, in_place="in_place")

        with pytest.raises(TypeError):
            sdk.deploy(config, canary_percent="10")

        with pytest.raises(ValueError):
            sdk.deploy(config, canary_percent=-1)

        with pytest.raises(ValueError):
            sdk.deploy(config, canary_percent=101)

        with pytest.raises(TypeError):
            sdk.deploy(config, max_surge_percent="10")

        with pytest.raises(ValueError):
            sdk.deploy(config, max_surge_percent=-1)

        with pytest.raises(ValueError):
            sdk.deploy(config, max_surge_percent=101)

    def test_basic(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

    def test_custom_image_from_containerfile(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        service_name = "test-service-name"
        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name=service_name,
            containerfile=TEST_CONTAINERFILE,
        )
        sdk.deploy(config)
        builds = fake_client.get_non_default_cluster_env_builds()
        assert len(builds) == 1
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name=service_name,
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=builds[0].id,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

    def test_custom_image_from_image_uri(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        image_uri = ImageURI.from_str("docker.io/user/my-custom-image:latest")
        build_id = fake_client.get_cluster_env_build_id_from_image_uri(
            image_uri=image_uri, ray_version="2.23.0",
        )

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            image_uri="docker.io/user/my-custom-image:latest",
            ray_version="2.23.0",
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=build_id,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

        sdk.deploy(config.options(image_uri="docker.io/user/does-not-exist:latest"))
        assert fake_client.rolled_out_model.build_id
        assert (
            fake_client.rolled_out_model.build_id != build_id
        )  # a new one should be created

    def test_custom_compute_config_name(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        fake_client.add_compute_config(
            ComputeTemplate(
                id="compute_id123",
                name="my-custom-compute-config",
                config=ComputeTemplateConfig(
                    cloud_id="test-cloud-id",
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )
        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            compute_config="my-custom-compute-config",
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id="compute_id123",
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

        with pytest.raises(
            ValueError, match="The compute config 'does-not-exist' does not exist."
        ):
            sdk.deploy(config.options(compute_config="does-not-exist"))

    def test_custom_compute_config_dict(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        fake_client.add_cloud(
            Cloud(
                id=str(uuid.uuid4()),
                name=TEST_COMPUTE_CONFIG_DICT["cloud"],
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )

        # Specify a compute config as a dictionary (anonymous).
        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            compute_config=TEST_COMPUTE_CONFIG_DICT,
        )
        sdk.deploy(config)
        anonymous_compute_config_id = fake_client.rolled_out_model.compute_config_id
        anonymous_compute_config = fake_client.get_compute_config(
            anonymous_compute_config_id
        )
        assert anonymous_compute_config.anonymous
        assert anonymous_compute_config.id == anonymous_compute_config_id
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=anonymous_compute_config_id,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

        # Test an invalid dict.
        with pytest.raises(
            TypeError,
            match=re.escape("__init__() got an unexpected keyword argument 'bad'"),
        ):
            config = ServiceConfig(
                applications=[{"import_path": "main:app"}],
                name="test-service-name",
                compute_config={"bad": "config"},
            )
            sdk.deploy(config)

    def test_reject_changed_cloud(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        fake_client.add_cloud(
            Cloud(
                id="test-cloud-id",
                name="test-cloud",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )
        fake_client.add_compute_config(
            ComputeTemplate(
                id="compute_id123",
                name="my-custom-compute-config",
                config=ComputeTemplateConfig(
                    cloud_id="test-cloud-id",
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        fake_client.add_cloud(
            Cloud(
                id="test-other-cloud-id",
                name="test-other-cloud",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )
        fake_client.add_compute_config(
            ComputeTemplate(
                id="compute_id456",
                name="my-other-compute-config",
                config=ComputeTemplateConfig(
                    cloud_id="test-other-cloud-id",
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            compute_config="my-custom-compute-config",
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id="compute_id123",
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

        with pytest.raises(
            ValueError,
            match="The cloud for a service cannot be changed once it's created. Service 'test-service-name' was created on cloud 'test-cloud', but the provided config is for cloud 'test-other-cloud'.",
        ):
            sdk.deploy(config.options(compute_config="my-other-compute-config"))

    def test_http_options(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            http_options={"request_timeout_s": 10.0},
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={
                "applications": config.applications,
                "http_options": {"request_timeout_s": 10.0},
            },
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

    def test_grpc_options_enabled(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            grpc_options={
                "grpc_servicer_functions": ["hello.world"],
                "service_names": ["hello."],
            },
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={
                "applications": config.applications,
                "grpc_options": {
                    "grpc_servicer_functions": ["hello.world"],
                    "service_names": ["hello."],
                },
            },
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(
                    grpc=GrpcProtocolConfig(enabled=True, service_names=["hello."])
                ),
            ),
        )

    def test_grpc_options_no_service_names(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            grpc_options={"grpc_servicer_functions": ["hello.world"]},
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={
                "applications": config.applications,
                "grpc_options": {"grpc_servicer_functions": ["hello.world"]},
            },
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

    def test_logging_config(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            logging_config={"log_level": "DEBUG"},
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={
                "applications": config.applications,
                "logging_config": {"log_level": "DEBUG"},
            },
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

    def test_in_place(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )

        sdk.deploy(config, in_place=False)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": [{"import_path": "main:app"}]},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

        config = config.options(applications=[{"import_path": "other:app"}])
        sdk.deploy(config, in_place=True)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": [{"import_path": "other:app"}]},
            rollout_strategy="IN_PLACE",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

        # Cluster-level options should be ignored, but application-level options should be applied.
        config = config.options(
            # These should be applied.
            applications=[{"import_path": "yet_other:app"}],
            working_dir="s3://remote.zip",
            requirements=["test"],
            logging_config={"log_level": "DEBUG"},
            # These should all be ignored.
            image_uri="docker.io/user/my-custom-image:latest",
            compute_config="foobar:123",
            ray_gcs_external_storage_config=RayGCSExternalStorageConfig(
                address="test.com:8000"
            ),
            query_auth_token_enabled=False,
        )
        sdk.deploy(config, in_place=True)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={
                "applications": [
                    {
                        "import_path": "yet_other:app",
                        "runtime_env": {
                            "working_dir": "s3://remote.zip",
                            "pip": ["test"],
                        },
                    }
                ],
                "logging_config": {"log_level": "DEBUG"},
            },
            rollout_strategy="IN_PLACE",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

    def test_in_place_service_does_not_exist(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )

        # in_place should fail if the service doesn't already exist.
        with pytest.raises(
            RuntimeError,
            match="in_place can only be used to update running services, but no service was found with name 'test-service-name'",
        ):
            sdk.deploy(config, in_place=True)

    def test_in_place_service_terminated(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(config)

        sdk.terminate(name="test-service-name")
        assert sdk.status(name="test-service-name").state == ServiceState.TERMINATED

        with pytest.raises(
            RuntimeError,
            match="in_place can only be used to update running services, but service 'test-service-name' is terminated",
        ):
            sdk.deploy(config, in_place=True)

    def test_in_place_during_rollout(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        # Service must exist before in_place is allowed.
        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(config, in_place=False)

        # Update the service to mimic an ongoing rollout (canary_version is populated).
        service = fake_client.get_service(
            name="test-service-name", cloud=None, project=None
        )
        service.canary_version = ProductionServiceV2VersionModel(
            version="canary",
            current_state=ExternalAPIServiceVersionState.STARTING,
            weight=0,
            build_id=service.primary_version.build_id,
            compute_config_id=service.primary_version.compute_config_id,
            ray_serve_config=service.primary_version.ray_serve_config,
            ray_gcs_external_storage_config=service.primary_version.ray_gcs_external_storage_config,
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )

        with pytest.raises(
            RuntimeError,
            match="in_place updates cannot be used while a rollout is in progress. Complete the rollout or roll back first.",
        ):
            sdk.deploy(config, in_place=True)

    def test_in_place_rejects_rollout_options(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        # Service must exist before in_place is allowed.
        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(config, in_place=False)

        # in_place does not support canary_percent or max_surge_percent.
        with pytest.raises(
            ValueError,
            match="canary_percent cannot be specified when doing an in_place update",
        ):
            sdk.deploy(config, in_place=True, canary_percent=50)

        with pytest.raises(
            ValueError,
            match="max_surge_percent cannot be specified when doing an in_place update",
        ):
            sdk.deploy(config, in_place=True, max_surge_percent=50)

    def test_canary_percent(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(config, canary_percent=50)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            canary_percent=50,
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

    def test_max_surge_percent(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(config, max_surge_percent=50)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            max_surge_percent=50,
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

    def test_disable_query_auth_token(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, logger, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            query_auth_token_enabled=False,
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=False),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

        assert not any("Bearer: " in info_msg for info_msg in logger.info_messages)

    def test_ray_gcs_external_storage_config(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            ray_gcs_external_storage_config=RayGCSExternalStorageConfig(
                address="test.com:8000"
            ),
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
            ray_gcs_external_storage_config=ExternalAPIRayGCSExternalStorageConfig(
                enable=True,
                address="test.com:8000",
                redis_certificate_path="/etc/ssl/certs/ca-certificates.crt",
            ),
        )

    def test_tracing_config(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            tracing_config=TracingConfig(
                enabled=True, exporter_import_path="path:exporter"
            ),
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
            tracing_config=ExternalAPITracingConfig(
                enabled=True, exporter_import_path="path:exporter", sampling_ratio=1,
            ),
        )

    def test_workspace_notification_is_sent(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.set_inside_workspace(True)

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(config)

        service_id = fake_client.get_service(
            "test-service-name", cloud=None, project=None
        ).id
        assert fake_client.sent_workspace_notifications == [
            WorkspaceNotification(
                body="Service 'test-service-name' deployed.",
                action=WorkspaceNotificationAction(
                    type="navigate-service", title="View Service", value=service_id,
                ),
            )
        ]

    @pytest.mark.parametrize(
        ("state", "expected_prefix"),
        [
            (None, "Starting new service"),
            (ServiceEventCurrentState.TERMINATED, "Restarting existing service"),
            (ServiceEventCurrentState.RUNNING, "Updating existing service"),
        ],
    )
    def test_existing_service_logs(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        state: Optional[ServiceEventCurrentState],
        expected_prefix: str,
    ):
        sdk, fake_client, logger, _ = service_sdk_with_fakes

        if state is not None:
            fake_client.update_service(
                ServiceModel(
                    id="test-service-id",
                    name="test-service-name",
                    current_state=state,
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                )
            )

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(config)

        assert any(
            info_log.startswith(expected_prefix) for info_log in logger.info_messages
        )

    def test_project(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        project = "test-project-name"
        project_id = fake_client.register_project_by_name(project)
        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            project=project,
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=project_id,
            build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ray_serve_config={"applications": config.applications},
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )

        with pytest.raises(
            ValueError,
            match="Cloud does not match from provided `cloud` and `compute_config`. Either pass one of these or ensure they match.",
        ):
            fake_client.add_cloud(
                Cloud(
                    id=str(uuid.uuid4()),
                    name=TEST_COMPUTE_CONFIG_DICT["cloud"],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
            )
            config = ServiceConfig(
                applications=[{"import_path": "main:app"}],
                name="test-service-name",
                compute_config=TEST_COMPUTE_CONFIG_DICT,
                cloud=fake_client.DEFAULT_CLOUD_NAME,
            )
            sdk.deploy(config)


class TestDeployWorkspaceDefaults:
    def test_name_defaults_to_workspace_name(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        # Happy path: workspace cluster name has the expected prefix.
        fake_client.set_inside_workspace(
            True,
            cluster_name=WORKSPACE_CLUSTER_NAME_PREFIX + "super-special-workspace-name",
        )
        config = ServiceConfig(applications=[{"import_path": "main:app"}])
        sdk.deploy(config)
        assert fake_client.rolled_out_model.name == "super-special-workspace-name"

        # Defensive path: workspace cluster name doesn't have the expected prefix.
        fake_client.set_inside_workspace(
            True, cluster_name="not-sure-how-this-happened"
        )
        config = ServiceConfig(applications=[{"import_path": "main:app"}])
        sdk.deploy(config)
        assert fake_client.rolled_out_model.name == "not-sure-how-this-happened"

    def test_pick_up_cluster_configs(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.set_inside_workspace(True)

        fake_client.add_compute_config(
            ComputeTemplate(
                id=fake_client.WORKSPACE_CLUSTER_COMPUTE_ID,
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.WORKSPACE_CLOUD_ID,
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            working_dir="s3://remote.zip",
        )
        sdk.deploy(config)
        assert fake_client.rolled_out_model == ApplyServiceModel(
            name="test-service-name",
            project_id=fake_client.WORKSPACE_PROJECT_ID,
            build_id=fake_client.WORKSPACE_CLUSTER_ENV_BUILD_ID,
            compute_config_id=fake_client.WORKSPACE_CLUSTER_COMPUTE_ID,
            ray_serve_config={
                "applications": [
                    {
                        "import_path": "main:app",
                        "runtime_env": {"working_dir": "s3://remote.zip",},
                    }
                ],
            },
            rollout_strategy="ROLLOUT",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=True),
                protocols=Protocols(grpc=GrpcProtocolConfig(enabled=False)),
            ),
        )


class TestOverrideApplicationRuntimeEnvs:
    @pytest.mark.parametrize(
        "new_py_modules", [None, [], ["A"], ["C", "D"]],
    )
    @pytest.mark.parametrize("existing_py_modules", [None, [], ["A"], ["B"]])
    def test_update_py_modules(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        new_py_modules: Optional[List[str]],
        existing_py_modules: Optional[List[str]],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        app: Dict[str, Any] = {"import_path": "main:app"}
        if existing_py_modules:
            app["runtime_env"] = {"py_modules": existing_py_modules}
        config = ServiceConfig(
            applications=[app], name="test-service-name", py_modules=new_py_modules,
        )
        sdk.deploy(config)
        applications = fake_client.rolled_out_model.ray_serve_config["applications"]

        if new_py_modules or existing_py_modules:
            if existing_py_modules is None:
                existing_py_modules = []
            if new_py_modules is None:
                new_py_modules = []
            uploaded_py_modules = [
                fake_client._upload_uri_mapping[local_dir]
                for local_dir in existing_py_modules + new_py_modules
            ]
        else:
            uploaded_py_modules = None

        if uploaded_py_modules:
            for app in applications:
                assert app["runtime_env"].get("py_modules") == uploaded_py_modules
        else:
            assert "py_modules" not in app.get("runtime_env", {})

    @pytest.mark.parametrize(
        "new_env_vars",
        [None, {}, {"A": "B"}, {"C": "D", "E": "F"}, {"K": "V1", "C": "D"}],
    )
    @pytest.mark.parametrize("existing_env_vars", [None, {}, {"K": "V"}, {"A": "B"}])
    @pytest.mark.parametrize("inside_workspace", [False, True])
    def test_update_env_vars(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        new_env_vars: Optional[Dict[str, str]],
        existing_env_vars: Optional[Dict[str, str]],
        inside_workspace: bool,
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.set_inside_workspace(inside_workspace)
        app: Dict[str, Any] = {"import_path": "main:app"}
        if existing_env_vars:
            app["runtime_env"] = {"env_vars": existing_env_vars}

        config = ServiceConfig(
            applications=[app], name="test-service-name", env_vars=new_env_vars,
        )
        sdk.deploy(config)
        applications = fake_client.rolled_out_model.ray_serve_config["applications"]
        for app in applications:
            if new_env_vars or existing_env_vars:
                if existing_env_vars is None:
                    existing_env_vars = {}
                if new_env_vars is None:
                    new_env_vars = {}
                assert app["runtime_env"].get("env_vars") == {
                    **existing_env_vars,
                    **new_env_vars,
                }
            else:
                assert "env_vars" not in app.get("runtime_env", {})

    @pytest.mark.parametrize("excludes_override", [None, ["override"]])
    @pytest.mark.parametrize(
        "working_dir_override", [None, "./some-local-path", "s3://some-remote-path.zip"]
    )
    @pytest.mark.parametrize("inside_workspace", [False, True])
    def test_override_working_dir_excludes(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        inside_workspace: bool,
        working_dir_override: Optional[str],
        excludes_override: Optional[List[str]],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.set_inside_workspace(inside_workspace)

        config = ServiceConfig(
            name="test-service-name",
            applications=[
                {"name": "no-runtime-env", "import_path": "main:app"},
                {
                    "name": "empty-runtime-env",
                    "import_path": "main:app",
                    "runtime_env": {},
                },
                {
                    "name": "other-working-dir",
                    "import_path": "main:app",
                    "runtime_env": {
                        "working_dir": "other-path/",
                        "excludes": ["existing"],
                    },
                },
                {
                    "name": "working-dir-none",
                    "import_path": "main:app",
                    # `None` should avoid workspace overriding.
                    "runtime_env": {"working_dir": None, "excludes": None},
                },
            ],
            working_dir=working_dir_override,
            excludes=excludes_override,
        )

        sdk.deploy(config)
        applications = fake_client.rolled_out_model.ray_serve_config["applications"]
        assert len(applications) == 4

        cloud_id = (
            FakeAnyscaleClient.WORKSPACE_CLOUD_ID
            if inside_workspace
            else FakeAnyscaleClient.DEFAULT_CLOUD_ID
        )
        cwd_uri = fake_client.upload_local_dir_to_cloud_storage(".", cloud_id=cloud_id)

        if working_dir_override is None and not inside_workspace:
            assert "working_dir" not in applications[0].get("runtime_env", {})
            assert "working_dir" not in applications[1]["runtime_env"]
            assert applications[2]["runtime_env"][
                "working_dir"
            ] == fake_client.upload_local_dir_to_cloud_storage(
                "other-path/", cloud_id=cloud_id,
            )
            assert applications[3]["runtime_env"]["working_dir"] is None
        elif working_dir_override is None and inside_workspace:
            assert applications[0]["runtime_env"]["working_dir"] == cwd_uri
            assert applications[1]["runtime_env"]["working_dir"] == cwd_uri
            assert applications[2]["runtime_env"][
                "working_dir"
            ] == fake_client.upload_local_dir_to_cloud_storage(
                "other-path/", cloud_id=cloud_id,
            )
            assert applications[3]["runtime_env"]["working_dir"] is None
        elif working_dir_override is not None and working_dir_override.startswith("s3"):
            assert applications[0]["runtime_env"]["working_dir"] == working_dir_override
            assert applications[1]["runtime_env"]["working_dir"] == working_dir_override
            assert applications[2]["runtime_env"]["working_dir"] == working_dir_override
            assert applications[3]["runtime_env"]["working_dir"] == working_dir_override
        else:
            override_uri = fake_client.upload_local_dir_to_cloud_storage(
                "./some-local-path", cloud_id=cloud_id
            )
            assert applications[0]["runtime_env"]["working_dir"] == override_uri
            assert applications[1]["runtime_env"]["working_dir"] == override_uri
            assert applications[2]["runtime_env"]["working_dir"] == override_uri
            assert applications[3]["runtime_env"]["working_dir"] == override_uri

        if excludes_override is None:
            assert "excludes" not in applications[0].get("runtime_env", {})
            assert "excludes" not in applications[1]["runtime_env"]
            assert applications[2]["runtime_env"]["excludes"] == ["existing"]
            assert applications[3]["runtime_env"]["excludes"] is None
        else:
            assert applications[0]["runtime_env"]["excludes"] == ["override"]
            assert applications[1]["runtime_env"]["excludes"] == ["override"]
            assert applications[2]["runtime_env"]["excludes"] == [
                "existing",
                "override",
            ]
            assert applications[3]["runtime_env"]["excludes"] == ["override"]

    @pytest.mark.parametrize(
        "enable_image_build_for_tracked_requirements", [False, True]
    )
    @pytest.mark.parametrize(
        "requirements_override",
        [None, MULTI_LINE_REQUIREMENTS.get_path(), ["override"]],
    )
    @pytest.mark.parametrize("workspace_tracking_enabled", [False, True])
    @pytest.mark.parametrize("inside_workspace", [False, True])
    def test_override_requirements(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        inside_workspace: bool,
        workspace_tracking_enabled: bool,
        requirements_override: Union[None, str, List[str]],
        enable_image_build_for_tracked_requirements: bool,
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.set_inside_workspace(
            inside_workspace,
            requirements_path=SINGLE_LINE_REQUIREMENTS.get_path()
            if workspace_tracking_enabled
            else None,
        )
        sdk._private_sdk._image_sdk._enable_image_build_for_tracked_requirements = (
            enable_image_build_for_tracked_requirements
        )

        config = ServiceConfig(
            name="test-service-name",
            applications=[
                {"name": "no-runtime-env", "import_path": "main:app",},
                {
                    "name": "empty-runtime-env",
                    "import_path": "main:app",
                    "runtime_env": {},
                },
                {
                    "name": "other-requirements-file",
                    "import_path": "main:app",
                    "runtime_env": {"pip": MULTI_LINE_REQUIREMENTS.get_path()},
                },
                {
                    "name": "other-requirements-list",
                    "import_path": "main:app",
                    "runtime_env": {"pip": ["testabc", "test123"]},
                },
                {
                    "name": "requirements-none",
                    "import_path": "main:app",
                    # `None` should disable overriding with workspace default.
                    "runtime_env": {"pip": None},
                },
            ],
            requirements=requirements_override,
        )

        sdk.deploy(config)
        if (
            inside_workspace
            and enable_image_build_for_tracked_requirements
            and workspace_tracking_enabled
        ):
            assert fake_client.rolled_out_model.build_id
        applications = fake_client.rolled_out_model.ray_serve_config["applications"]
        assert len(applications) == 5

        if isinstance(requirements_override, str):
            # Override with a file.
            pass
        elif isinstance(requirements_override, list):
            # Override with a list.
            pass
        elif (
            inside_workspace
            and workspace_tracking_enabled
            and not enable_image_build_for_tracked_requirements
        ):
            # Workspace default.
            expected_workspace_pip = SINGLE_LINE_REQUIREMENTS.expected_pip_list
            assert applications[0]["runtime_env"]["pip"] == expected_workspace_pip
            assert applications[1]["runtime_env"]["pip"] == expected_workspace_pip
            assert (
                applications[2]["runtime_env"]["pip"]
                == MULTI_LINE_REQUIREMENTS.expected_pip_list
            )
            assert applications[3]["runtime_env"]["pip"] == ["testabc", "test123"]
            assert applications[4]["runtime_env"]["pip"] is None
        else:
            # No overrides.
            assert "pip" not in applications[0].get("runtime_env", {})
            assert "pip" not in applications[1]["runtime_env"]
            assert (
                applications[2]["runtime_env"]["pip"]
                == MULTI_LINE_REQUIREMENTS.expected_pip_list
            )
            assert applications[3]["runtime_env"]["pip"] == ["testabc", "test123"]
            assert applications[4]["runtime_env"]["pip"] is None


class TestDeployUploadDirs:
    @pytest.mark.parametrize("inside_workspace", [False, True])
    def test_upload_basic_working_dir(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        inside_workspace: bool,
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.set_inside_workspace(inside_workspace)

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            working_dir=".",
        )
        original_config = copy.deepcopy(config)

        sdk.deploy(config)
        # The original config should not be modified.
        assert config == original_config

        [application] = fake_client.rolled_out_model.ray_serve_config["applications"]

        # Check that the correct cloud_id was used for the upload.
        expected_cloud_id = (
            FakeAnyscaleClient.WORKSPACE_CLOUD_ID
            if inside_workspace
            else FakeAnyscaleClient.DEFAULT_CLOUD_ID
        )
        assert application["runtime_env"]["working_dir"].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(cloud_id=expected_cloud_id)
        )

    def test_upload_uses_cloud_from_compute_config(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.add_compute_config(
            ComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
                config=ComputeTemplateConfig(
                    cloud_id="compute-config-cloud-id",
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
            )
        )

        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            working_dir=".",
            compute_config="fake-compute-config",
        )
        sdk.deploy(config)

        [application] = fake_client.rolled_out_model.ray_serve_config["applications"]

        # Check that the correct cloud_id was used for the upload.
        expected_cloud_id = "compute-config-cloud-id"
        assert application["runtime_env"]["working_dir"].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(cloud_id=expected_cloud_id)
        )

    def test_upload_with_no_local_dirs(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        """Configs should be left unchanged if there are no local dirs."""
        sdk, fake_client, _, _ = service_sdk_with_fakes

        basic_config = ServiceConfig(
            applications=[{"import_path": "main:app"}], name="test-service-name"
        )
        sdk.deploy(basic_config)
        assert (
            fake_client.rolled_out_model.ray_serve_config["applications"]
            == basic_config.applications
        )

        config_with_requirements = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            requirements=["pip-install-test"],
        )
        sdk.deploy(config_with_requirements)
        assert fake_client.rolled_out_model.ray_serve_config["applications"] == [
            {"import_path": "main:app", "runtime_env": {"pip": ["pip-install-test"]},},
        ]

        complex_config = ServiceConfig(
            name="test-service-name",
            applications=[
                {"name": "app1", "import_path": "main:app",},
                {
                    "name": "app2",
                    "import_path": "main:app",
                    "runtime_env": {"env_vars": {"foo": "bar",},},
                },
            ],
        )
        sdk.deploy(complex_config)
        assert (
            fake_client.rolled_out_model.ray_serve_config["applications"]
            == complex_config.applications
        )

    def test_no_upload_remote_working_dir(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        config = ServiceConfig(
            applications=[{"import_path": "main:app"}],
            name="test-service-name",
            working_dir="s3://some-remote-uri.zip",
        )

        sdk.deploy(config)
        applications = fake_client.rolled_out_model.ray_serve_config["applications"]
        assert len(applications) == 1
        assert (
            applications[0]["runtime_env"]["working_dir"] == "s3://some-remote-uri.zip"
        )

    def test_upload_local_py_modules(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        config = ServiceConfig(
            name="test-service-name",
            applications=[
                {
                    "import_path": "main:app",
                    "runtime_env": {
                        "py_modules": [
                            # Should be left alone.
                            "s3://some-remote-uri.zip",
                            # Should be uploaded.
                            "local-path",
                        ],
                    },
                },
            ],
        )
        sdk.deploy(config)
        applications = fake_client.rolled_out_model.ray_serve_config["applications"]
        assert len(applications) == 1
        assert (
            applications[0]["runtime_env"]["py_modules"][0]
            == "s3://some-remote-uri.zip"
        )
        assert applications[0]["runtime_env"]["py_modules"][1].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(
                cloud_id=FakeAnyscaleClient.DEFAULT_CLOUD_ID
            )
        )

    def test_upload_caching(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        """The same directory should only by uploaded once."""
        sdk, fake_client, _, _ = service_sdk_with_fakes
        config = ServiceConfig(
            name="test-service-name",
            applications=[
                {
                    "name": "app1",
                    "import_path": "main:app",
                    "runtime_env": {"working_dir": ".",},
                },
                {
                    "name": "app2",
                    "import_path": "main:app",
                    "runtime_env": {
                        "working_dir": ".",
                        "py_modules": [".", "other-dir",],
                    },
                },
            ],
        )
        sdk.deploy(config)
        applications = fake_client.rolled_out_model.ray_serve_config["applications"]
        assert len(applications) == 2
        assert applications[0]["runtime_env"]["working_dir"].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(
                cloud_id=FakeAnyscaleClient.DEFAULT_CLOUD_ID
            )
        )
        common_uri = applications[0]["runtime_env"]["working_dir"]

        assert applications[1]["runtime_env"]["working_dir"] == common_uri
        assert applications[1]["runtime_env"]["py_modules"][0] == common_uri
        assert applications[1]["runtime_env"]["py_modules"][1] != common_uri
        assert applications[1]["runtime_env"]["py_modules"][1].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(
                cloud_id=FakeAnyscaleClient.DEFAULT_CLOUD_ID
            )
        )


class TestLoadRequirementsFiles:
    @pytest.mark.parametrize("requirements", [None, *TEST_REQUIREMENTS_FILES])
    @pytest.mark.parametrize("in_applications", [False, True])
    def test_override_requirements_file(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        requirements: Optional[RequirementsFile],
        in_applications: bool,
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        applications: List[Dict[str, Any]] = [
            {"name": "no_runtime_env", "import_path": "main:app"},
            {
                "name": "empty_runtime_env",
                "import_path": "main:app",
                "runtime_env": {},
            },
            {
                "name": "has_runtime_env",
                "import_path": "main:app",
                "runtime_env": {
                    "env_vars": {"abc": "123"},
                    "working_dir": "s3://somewhere.zip",
                },
            },
        ]
        if in_applications and requirements is not None:
            for application in applications:
                assert isinstance(application, dict)
                runtime_env = application.get("runtime_env", {})
                runtime_env["pip"] = requirements.get_path()
                application["runtime_env"] = runtime_env

            config = ServiceConfig(name="test-service-name", applications=applications)
        else:
            config = ServiceConfig(
                name="test-service-name",
                applications=applications,
                requirements=requirements.get_path() if requirements else None,
            )

        if requirements is not None and requirements.expected_pip_list is None:
            with pytest.raises(FileNotFoundError):
                sdk.deploy(config)

            return

        sdk.deploy(config)
        deployed_ray_serve_config = fake_client.rolled_out_model.ray_serve_config
        if requirements is None:
            assert deployed_ray_serve_config["applications"] == config.applications
        else:
            assert len(deployed_ray_serve_config["applications"]) == 3
            assert deployed_ray_serve_config["applications"][0] == {
                "name": "no_runtime_env",
                "import_path": "main:app",
                "runtime_env": {"pip": requirements.expected_pip_list},
            }

            assert deployed_ray_serve_config["applications"][1] == {
                "name": "empty_runtime_env",
                "import_path": "main:app",
                "runtime_env": {"pip": requirements.expected_pip_list},
            }
            assert deployed_ray_serve_config["applications"][2] == {
                "name": "has_runtime_env",
                "import_path": "main:app",
                "runtime_env": {
                    "env_vars": {"abc": "123"},
                    "working_dir": "s3://somewhere.zip",
                    "pip": requirements.expected_pip_list,
                },
            }


TEST_SERVICE_ID = "fake-service-id"
TEST_SERVICE_NAME = "fake-service-name"
TEST_QUERY_URL = "fake-query-url"

TEST_IMAGE_URI = ImageURI.from_str("user/fake-build:3")
TEST_CLUSTER_ENV_BUILD_ID = "fake-build-id"

TEST_COMPUTE_CONFIG_NAME = "fake-compute-config"
TEST_COMPUTE_CONFIG_ID = "fake-compute-config-id"

TEST_APPLICATIONS = [{"import_path": "main:app"}]
TEST_HTTP_OPTIONS = {
    "request_timeout_s": 10.0,
}
TEST_GRPC_OPTIONS = {
    "grpc_servicer_functions": ["hello.world"],
}

DEFAULT_EXPECTED_STATUS = ServiceStatus(
    id=TEST_SERVICE_ID,
    name=TEST_SERVICE_NAME,
    state=ServiceState.TERMINATED,
    query_url=TEST_QUERY_URL,
    query_auth_token=None,
    primary_version=None,
    canary_version=None,
)


def _expected_status(**kwargs) -> ServiceStatus:
    return ServiceStatus(**{**DEFAULT_EXPECTED_STATUS.to_dict(), **kwargs})


def _service_version_model(
    id: str,  # noqa: A002
    weight: int,
    *,
    state: ExternalAPIServiceVersionState = ExternalAPIServiceVersionState.RUNNING,
    build_id: str = TEST_CLUSTER_ENV_BUILD_ID,
    compute_config_id: str = TEST_COMPUTE_CONFIG_ID,
    applications: Optional[List[Dict[str, Any]]] = None,
    http_options: Optional[Dict[str, Any]] = None,
    grpc_options: Optional[Dict[str, Any]] = None,
    logging_config: Optional[Dict[str, Any]] = None,
    ray_gcs_external_storage_config: Optional[Dict[str, Any]] = None,
    tracing_config: Optional[Dict[str, Any]] = None,
) -> ProductionServiceV2VersionModel:
    ray_serve_config: Dict[str, Any] = {
        "applications": applications or TEST_APPLICATIONS
    }
    if http_options:
        ray_serve_config["http_options"] = http_options
    if grpc_options:
        ray_serve_config["grpc_options"] = grpc_options
    if logging_config:
        ray_serve_config["logging_config"] = logging_config

    return ProductionServiceV2VersionModel(
        id=id,
        current_state=state,
        weight=weight,
        build_id=build_id,
        compute_config_id=compute_config_id,
        ray_serve_config=ray_serve_config,
        ray_gcs_external_storage_config=ExternalAPIRayGCSExternalStorageConfig(
            **ray_gcs_external_storage_config
        )
        if ray_gcs_external_storage_config
        else None,
        tracing_config=ExternalAPITracingConfig(**tracing_config)
        if tracing_config
        else None,
        local_vars_configuration=OPENAPI_NO_VALIDATION,
    )


@dataclass
class ServiceStatusTestCase:
    id: str = TEST_SERVICE_ID
    name: str = TEST_SERVICE_NAME
    state: ServiceEventCurrentState = ServiceEventCurrentState.TERMINATED
    base_url: str = TEST_QUERY_URL
    auth_token: Optional[str] = None
    primary_version: Optional[ProductionServiceV2VersionModel] = None
    canary_version: Optional[ProductionServiceV2VersionModel] = None

    expected_status: ServiceStatus = DEFAULT_EXPECTED_STATUS

    def service_model(self) -> ServiceModel:
        return ServiceModel(
            id=self.id,
            name=self.name,
            current_state=self.state,
            base_url=self.base_url,
            auth_token=self.auth_token,
            primary_version=self.primary_version,
            canary_version=self.canary_version,
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )


SERVICE_STATUS_TEST_CASES = [
    ServiceStatusTestCase(),
    ServiceStatusTestCase(expected_status=DEFAULT_EXPECTED_STATUS),
    # Test that states are mapped correctly.
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.STARTING,
        expected_status=_expected_status(state=ServiceState.STARTING),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.RUNNING,
        expected_status=_expected_status(state=ServiceState.RUNNING),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.ROLLING_OUT,
        expected_status=_expected_status(state=ServiceState.ROLLING_OUT),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.ROLLING_BACK,
        expected_status=_expected_status(state=ServiceState.ROLLING_BACK),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.UPDATING,
        expected_status=_expected_status(state=ServiceState.UPDATING),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.TERMINATING,
        expected_status=_expected_status(state=ServiceState.TERMINATING),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.TERMINATED,
        expected_status=_expected_status(state=ServiceState.TERMINATED),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.UNHEALTHY,
        expected_status=_expected_status(state=ServiceState.UNHEALTHY),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.SYSTEM_FAILURE,
        expected_status=_expected_status(state=ServiceState.SYSTEM_FAILURE),
    ),
    # Test that unrecognized states are mapped to ServiceState.UNKNOWN.
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.USER_ERROR_FAILURE,
        expected_status=_expected_status(state=ServiceState.UNKNOWN),
    ),
    ServiceStatusTestCase(
        auth_token="test-token-abc123",
        expected_status=_expected_status(query_auth_token="test-token-abc123"),
    ),
    # Test that primary and canary versions are mapped correctly.
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.RUNNING,
        primary_version=_service_version_model("primary", 100),
        expected_status=_expected_status(
            state=ServiceState.RUNNING,
            primary_version=ServiceVersionStatus(
                id="primary",
                state=ServiceVersionState.RUNNING,
                weight=100,
                config=ServiceConfig(
                    name=TEST_SERVICE_NAME,
                    applications=TEST_APPLICATIONS,
                    image_uri=TEST_IMAGE_URI.image_uri,
                    compute_config=TEST_COMPUTE_CONFIG_NAME + ":1",
                    query_auth_token_enabled=False,
                ),
            ),
        ),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.RUNNING,
        primary_version=_service_version_model("primary", 50),
        canary_version=_service_version_model(
            "canary",
            0,
            http_options=TEST_HTTP_OPTIONS,
            grpc_options=TEST_GRPC_OPTIONS,
            state=ExternalAPIServiceVersionState.STARTING,
        ),
        expected_status=_expected_status(
            state=ServiceState.RUNNING,
            primary_version=ServiceVersionStatus(
                id="primary",
                state=ServiceVersionState.RUNNING,
                weight=50,
                config=ServiceConfig(
                    name=TEST_SERVICE_NAME,
                    applications=TEST_APPLICATIONS,
                    image_uri=TEST_IMAGE_URI.image_uri,
                    compute_config=TEST_COMPUTE_CONFIG_NAME + ":1",
                    query_auth_token_enabled=False,
                ),
            ),
            canary_version=ServiceVersionStatus(
                id="canary",
                state=ServiceVersionState.STARTING,
                weight=0,
                config=ServiceConfig(
                    name=TEST_SERVICE_NAME,
                    applications=TEST_APPLICATIONS,
                    image_uri=TEST_IMAGE_URI.image_uri,
                    compute_config=TEST_COMPUTE_CONFIG_NAME + ":1",
                    query_auth_token_enabled=False,
                    http_options=TEST_HTTP_OPTIONS,
                    grpc_options=TEST_GRPC_OPTIONS,
                ),
            ),
        ),
    ),
    # Test that when the service is TERMINATED, primary/canary versions are ignored.
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.TERMINATED,
        primary_version=_service_version_model("primary", 100),
        expected_status=_expected_status(state=ServiceState.TERMINATED),
    ),
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.TERMINATED,
        canary_version=_service_version_model("canary", 100),
        expected_status=_expected_status(state=ServiceState.TERMINATED),
    ),
    # Test that tracing_config is mapped correctly
    ServiceStatusTestCase(
        state=ServiceEventCurrentState.RUNNING,
        primary_version=_service_version_model(
            "primary",
            100,
            tracing_config={
                "exporter_import_path": "path:exporter",
                "sampling_ratio": 1.0,
                "enabled": True,
            },
        ),
        expected_status=_expected_status(
            state=ServiceState.RUNNING,
            primary_version=ServiceVersionStatus(
                id="primary",
                state=ServiceVersionState.RUNNING,
                weight=100,
                config=ServiceConfig(
                    name=TEST_SERVICE_NAME,
                    applications=TEST_APPLICATIONS,
                    image_uri=TEST_IMAGE_URI.image_uri,
                    compute_config=TEST_COMPUTE_CONFIG_NAME + ":1",
                    query_auth_token_enabled=False,
                    tracing_config=TracingConfig(
                        exporter_import_path="path:exporter",
                        sampling_ratio=1.0,
                        enabled=True,
                    ),
                ),
            ),
        ),
    ),
]


class TestStatus:
    def test_service_not_found(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        with pytest.raises(
            RuntimeError, match="Service with name 'test-service-name' was not found."
        ):
            sdk.status("test-service-name")

    @pytest.mark.parametrize("test_case", SERVICE_STATUS_TEST_CASES)
    def test_build_status_from_model(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        test_case: ServiceStatusTestCase,
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.add_build(
            ClusterEnvironmentBuild(
                id=TEST_CLUSTER_ENV_BUILD_ID,
                docker_image_name=str(TEST_IMAGE_URI),
                status=ClusterEnvironmentBuildStatus.SUCCEEDED,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        fake_client.add_compute_config(
            ComputeTemplate(
                id=TEST_COMPUTE_CONFIG_ID,
                name=TEST_COMPUTE_CONFIG_NAME,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )
        service_model = test_case.service_model()
        fake_client.update_service(service_model)

        status: ServiceStatus = sdk.status(name=service_model.name)
        assert (
            status == test_case.expected_status
        ), f"Expected: {test_case.expected_status}\nActual: {status}"

    def test_failed_to_get_cluster_env_build_image_uri(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        test_case = ServiceStatusTestCase(
            state=ServiceEventCurrentState.RUNNING,
            primary_version=_service_version_model("primary", 100),
        )
        service_model = test_case.service_model()
        fake_client.update_service(service_model)
        fake_client.add_compute_config(
            ComputeTemplate(
                id=TEST_COMPUTE_CONFIG_ID,
                name=TEST_COMPUTE_CONFIG_NAME,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        with pytest.raises(
            RuntimeError, match="Failed to get image URI for ID fake-build-id"
        ):
            sdk.status(name=service_model.name)

    def test_failed_to_get_compute_config_name(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        test_case = ServiceStatusTestCase(
            state=ServiceEventCurrentState.RUNNING,
            primary_version=_service_version_model("primary", 100),
        )
        service_model = test_case.service_model()
        fake_client.update_service(service_model)
        fake_client.add_build(
            ClusterEnvironmentBuild(
                id=TEST_CLUSTER_ENV_BUILD_ID,
                docker_image_name=str(TEST_IMAGE_URI),
                status=ClusterEnvironmentBuildStatus.SUCCEEDED,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        with pytest.raises(RuntimeError, match="Failed to get compute config"):
            sdk.status(name=service_model.name)

    def test_anonymous_compute_config_inlined(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.add_build(
            ClusterEnvironmentBuild(
                id=TEST_CLUSTER_ENV_BUILD_ID,
                docker_image_name=str(TEST_IMAGE_URI),
                status=ClusterEnvironmentBuildStatus.SUCCEEDED,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )
        fake_client.add_compute_config(
            ComputeTemplate(
                anonymous=True,
                id="anonymous-compute-config-id",
                name="anonymous-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    head_node_type=ComputeNodeType(
                        name="head-node",
                        instance_type="head-node-instance-type",
                        flags={},
                    ),
                    worker_node_types=[],
                    auto_select_worker_config=True,
                    flags={},
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        test_case = ServiceStatusTestCase(
            state=ServiceEventCurrentState.RUNNING,
            primary_version=_service_version_model(
                "primary", 100, compute_config_id="anonymous-compute-config-id"
            ),
            expected_status=_expected_status(
                state=ServiceState.RUNNING,
                primary_version=ServiceVersionStatus(
                    id="primary",
                    state=ServiceVersionState.RUNNING,
                    weight=100,
                    config=ServiceConfig(
                        name=TEST_SERVICE_NAME,
                        applications=TEST_APPLICATIONS,
                        image_uri=TEST_IMAGE_URI.image_uri,
                        compute_config=ComputeConfig(
                            cloud=fake_client.DEFAULT_CLOUD_NAME,
                            head_node=HeadNodeConfig(
                                instance_type="head-node-instance-type",
                            ),
                            flags={},
                        ),
                        cloud=fake_client.DEFAULT_CLOUD_NAME,
                        query_auth_token_enabled=False,
                    ),
                ),
            ),
        )
        service_model = test_case.service_model()
        fake_client.update_service(service_model)

        status: ServiceStatus = sdk.status(name=service_model.name)
        assert (
            status == test_case.expected_status
        ), f"Expected: {test_case.expected_status}\nActual: {status}"

    def test_ray_gcs_external_storage_config(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.add_build(
            ClusterEnvironmentBuild(
                id=TEST_CLUSTER_ENV_BUILD_ID,
                docker_image_name=str(TEST_IMAGE_URI),
                status=ClusterEnvironmentBuildStatus.SUCCEEDED,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )
        fake_client.add_compute_config(
            ComputeTemplate(
                id="compute-config-id",
                name="compute-config-name",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        test_case = ServiceStatusTestCase(
            state=ServiceEventCurrentState.RUNNING,
            primary_version=_service_version_model(
                "primary",
                100,
                compute_config_id="compute-config-id",
                ray_gcs_external_storage_config={"address": "host:1234"},
            ),
            expected_status=_expected_status(
                state=ServiceState.RUNNING,
                primary_version=ServiceVersionStatus(
                    id="primary",
                    state=ServiceVersionState.RUNNING,
                    weight=100,
                    config=ServiceConfig(
                        name=TEST_SERVICE_NAME,
                        applications=TEST_APPLICATIONS,
                        image_uri=TEST_IMAGE_URI.image_uri,
                        compute_config="compute-config-name:1",
                        query_auth_token_enabled=False,
                        ray_gcs_external_storage_config=RayGCSExternalStorageConfig(
                            enabled=True,
                            address="host:1234",
                            certificate_path="/etc/ssl/certs/ca-certificates.crt",
                        ),
                    ),
                ),
            ),
        )
        service_model = test_case.service_model()
        fake_client.update_service(service_model)

        status: ServiceStatus = sdk.status(name=service_model.name)
        assert (
            status == test_case.expected_status
        ), f"Expected: {test_case.expected_status}\nActual: {status}"

    def test_logging_config(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.add_build(
            ClusterEnvironmentBuild(
                id=TEST_CLUSTER_ENV_BUILD_ID,
                docker_image_name=str(TEST_IMAGE_URI),
                status=ClusterEnvironmentBuildStatus.SUCCEEDED,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        fake_client.add_compute_config(
            ComputeTemplate(
                id="compute-config-id",
                name="compute-config-name",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        test_case = ServiceStatusTestCase(
            state=ServiceEventCurrentState.RUNNING,
            primary_version=_service_version_model(
                "primary",
                100,
                compute_config_id="compute-config-id",
                logging_config={"log_level": "DEBUG"},
            ),
            expected_status=_expected_status(
                state=ServiceState.RUNNING,
                primary_version=ServiceVersionStatus(
                    id="primary",
                    state=ServiceVersionState.RUNNING,
                    weight=100,
                    config=ServiceConfig(
                        name=TEST_SERVICE_NAME,
                        applications=TEST_APPLICATIONS,
                        image_uri=TEST_IMAGE_URI.image_uri,
                        compute_config="compute-config-name:1",
                        query_auth_token_enabled=False,
                        logging_config={"log_level": "DEBUG"},
                    ),
                ),
            ),
        )
        service_model = test_case.service_model()
        fake_client.update_service(service_model)

        status: ServiceStatus = sdk.status(name=service_model.name)
        assert (
            status == test_case.expected_status
        ), f"Expected: {test_case.expected_status}\nActual: {status}"

    def test_versioned_compute_config_name(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        """Verify that versioned compute config names are constructed properly."""
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.add_build(
            ClusterEnvironmentBuild(
                id=TEST_CLUSTER_ENV_BUILD_ID,
                docker_image_name=str(TEST_IMAGE_URI),
                status=ClusterEnvironmentBuildStatus.SUCCEEDED,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        # Version 1.
        fake_client.add_compute_config(
            ComputeTemplate(
                id="v1-compute-config-id",
                name="test-compute-config",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )
        test_case = ServiceStatusTestCase(
            state=ServiceEventCurrentState.RUNNING,
            primary_version=_service_version_model(
                "primary", 100, compute_config_id="v1-compute-config-id",
            ),
            expected_status=_expected_status(
                state=ServiceState.RUNNING,
                primary_version=ServiceVersionStatus(
                    id="primary",
                    state=ServiceVersionState.RUNNING,
                    weight=100,
                    config=ServiceConfig(
                        name=TEST_SERVICE_NAME,
                        applications=TEST_APPLICATIONS,
                        image_uri=TEST_IMAGE_URI.image_uri,
                        compute_config="test-compute-config:1",
                        query_auth_token_enabled=False,
                    ),
                ),
            ),
        )
        service_model = test_case.service_model()
        fake_client.update_service(service_model)

        v1_status: ServiceStatus = sdk.status(name=service_model.name)
        assert (
            v1_status == test_case.expected_status
        ), f"Expected: {test_case.expected_status}\nActual: {v1_status}"

        # Version 2.
        fake_client.add_compute_config(
            ComputeTemplate(
                id="v2-compute-config-id",
                name="test-compute-config",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )
        test_case = ServiceStatusTestCase(
            state=ServiceEventCurrentState.RUNNING,
            primary_version=_service_version_model(
                "primary", 100, compute_config_id="v2-compute-config-id"
            ),
            expected_status=_expected_status(
                state=ServiceState.RUNNING,
                primary_version=ServiceVersionStatus(
                    id="primary",
                    state=ServiceVersionState.RUNNING,
                    weight=100,
                    config=ServiceConfig(
                        name=TEST_SERVICE_NAME,
                        applications=TEST_APPLICATIONS,
                        image_uri=TEST_IMAGE_URI.image_uri,
                        compute_config="test-compute-config:2",
                        query_auth_token_enabled=False,
                    ),
                ),
            ),
        )
        service_model = test_case.service_model()
        fake_client.update_service(service_model)

        v2_status: ServiceStatus = sdk.status(name=service_model.name)
        assert (
            v2_status == test_case.expected_status
        ), f"Expected: {test_case.expected_status}\nActual: {v2_status}"

    def test_status_with_projects(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        project = "test-project-name"
        name = "test-service-name"
        cloud = "fake-default-cloud"

        fake_client.register_project_by_name(
            name=fake_client.DEFAULT_PROJECT_NAME,
            project_id=fake_client.DEFAULT_PROJECT_ID,
        )
        fake_client.register_project_by_name(name=project, cloud=cloud)

        service_id = sdk.deploy(
            ServiceConfig(
                applications=[{"import_path": "main:app"}],
                name=name,
                cloud=cloud,
                project=project,
            )
        )

        # deploy another service with the same name, under the default project
        sdk.deploy(
            ServiceConfig(
                applications=[{"import_path": "main:app"}], name=name, cloud=cloud
            )
        )

        # ensure that the project filter works
        status: ServiceStatus = sdk.status(name=name, cloud=cloud, project=project)
        assert status.id == service_id
        assert status.primary_version.config.project == project  # type: ignore


class TestRollback:
    def test_not_found(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        with pytest.raises(
            RuntimeError, match="Service with name 'does-not-exist' was not found."
        ):
            sdk.rollback(name="does-not-exist")

    def test_basic(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        service_id = "test-service-id"
        fake_client.update_service(
            ServiceModel(
                id=service_id,
                name="test-name",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        sdk.rollback(name="test-name")
        assert fake_client.rolled_back_service == ("test-service-id", None)

    def test_max_surge_percent(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        service_id = "test-service-id"
        fake_client.update_service(
            ServiceModel(
                id=service_id,
                name="test-name",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        sdk.rollback(name="test-name", max_surge_percent=50)
        assert fake_client.rolled_back_service == ("test-service-id", 50)

    def test_no_name_passed_outside_workspace(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        with pytest.raises(
            ValueError,
            match="A service name must be provided when running outside of a workspace.",
        ):
            sdk.rollback()

    def test_default_name_in_workspace(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.set_inside_workspace(
            True,
            cluster_name=WORKSPACE_CLUSTER_NAME_PREFIX + "default-name-from-workspace",
        )

        service_id = "test-service-id"
        fake_client.update_service(
            ServiceModel(
                id=service_id,
                name="default-name-from-workspace",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        sdk.rollback()
        assert fake_client.rolled_back_service == ("test-service-id", None)


class TestTerminate:
    def test_not_found(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        with pytest.raises(
            RuntimeError, match="Service with name 'does-not-exist' was not found."
        ):
            sdk.terminate("does-not-exist")

    def test_basic(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        service_id = "test-service-id"
        fake_client.update_service(
            ServiceModel(
                id=service_id,
                name="test-name",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        sdk.terminate("test-name")
        assert fake_client.terminated_service == "test-service-id"

    def test_no_name_passed_outside_workspace(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes

        with pytest.raises(
            ValueError,
            match="A service name must be provided when running outside of a workspace.",
        ):
            sdk.terminate()

    def test_default_name_in_workspace(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        fake_client.set_inside_workspace(
            True,
            cluster_name=WORKSPACE_CLUSTER_NAME_PREFIX + "default-name-from-workspace",
        )

        service_id = "test-service-id"
        fake_client.update_service(
            ServiceModel(
                id=service_id,
                name="default-name-from-workspace",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        sdk.terminate()
        assert fake_client.terminated_service == "test-service-id"


class TestWait:
    def test_service_not_found(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = service_sdk_with_fakes
        with pytest.raises(
            RuntimeError, match="Service with name 'test-service-name' was not found."
        ):
            sdk.wait(
                "test-service-name",
                state=ServiceState.RUNNING,
                timeout_s=60,
                _interval_s=1,
            )

    def test_succeeds_immediately(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, logger, _ = service_sdk_with_fakes
        fake_client.update_service(
            ServiceModel(
                id=str(uuid.uuid4()),
                name="test-service-name",
                current_state=ServiceEventCurrentState.RUNNING,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )

        sdk.wait(
            "test-service-name",
            state=ServiceState.RUNNING,
            timeout_s=60,
            _interval_s=1,
        )
        assert any(
            "Service 'test-service-name' reached target state, exiting" in info_log
            for info_log in logger.info_messages
        )

    def test_times_out(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, logger, _ = service_sdk_with_fakes
        fake_client.update_service(
            ServiceModel(
                id=str(uuid.uuid4()),
                name="test-service-name",
                current_state=ServiceEventCurrentState.ROLLING_OUT,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )

        with pytest.raises(
            TimeoutError,
            match="Service 'test-service-name' did not reach target state RUNNING within 60s. Last seen state: ROLLING_OUT.",
        ):
            sdk.wait(
                "test-service-name",
                state=ServiceState.RUNNING,
                timeout_s=60,
                _interval_s=1,
            )

    def test_start_sequence(
        self,
        service_sdk_with_fakes: Tuple[
            ServiceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, logger, timer = service_sdk_with_fakes
        base_model = ServiceModel(
            id=str(uuid.uuid4()),
            name="test-service-name",
            current_state=ServiceEventCurrentState.STARTING,
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )
        fake_client.update_service(base_model)

        def start_sequence(i: int):
            model = copy.deepcopy(base_model)
            if i == 0:
                model.current_state = ServiceEventCurrentState.STARTING
            elif i == 1:
                model.current_state = ServiceEventCurrentState.UPDATING
            else:
                model.current_state = ServiceEventCurrentState.RUNNING

            fake_client.update_service(model)

        timer.set_on_poll_iteration(start_sequence)

        sdk.wait(
            "test-service-name",
            state=ServiceState.RUNNING,
            timeout_s=60,
            _interval_s=1,
        )

        assert any(
            "Service 'test-service-name' transitioned from STARTING to UPDATING"
            in info_log
            for info_log in logger.info_messages
        )
        assert any(
            "Service 'test-service-name' transitioned from UPDATING to RUNNING"
            in info_log
            for info_log in logger.info_messages
        )
        assert any(
            "Service 'test-service-name' reached target state, exiting" in info_log
            for info_log in logger.info_messages
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class AggregatedInstanceUsageWithCostModel(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'date': 'date',
        'organization_id': 'str',
        'organization_name': 'str',
        'cloud_hosting_type': 'CloudHostingType',
        'cloud_id': 'str',
        'cloud_name': 'str',
        'project_id': 'str',
        'project_name': 'str',
        'user_id': 'str',
        'user_email': 'str',
        'cluster_id': 'str',
        'workspace_id': 'str',
        'job_id': 'str',
        'job_queue_id': 'str',
        'service_id': 'str',
        'instance_type': 'str',
        'instance_seconds': 'float',
        'cloud_provider': 'CloudProviders',
        'anyscale_credits': 'float'
    }

    attribute_map = {
        'date': 'date',
        'organization_id': 'organization_id',
        'organization_name': 'organization_name',
        'cloud_hosting_type': 'cloud_hosting_type',
        'cloud_id': 'cloud_id',
        'cloud_name': 'cloud_name',
        'project_id': 'project_id',
        'project_name': 'project_name',
        'user_id': 'user_id',
        'user_email': 'user_email',
        'cluster_id': 'cluster_id',
        'workspace_id': 'workspace_id',
        'job_id': 'job_id',
        'job_queue_id': 'job_queue_id',
        'service_id': 'service_id',
        'instance_type': 'instance_type',
        'instance_seconds': 'instance_seconds',
        'cloud_provider': 'cloud_provider',
        'anyscale_credits': 'anyscale_credits'
    }

    def __init__(self, date=None, organization_id=None, organization_name=None, cloud_hosting_type=None, cloud_id=None, cloud_name=None, project_id=None, project_name=None, user_id=None, user_email=None, cluster_id=None, workspace_id=None, job_id=None, job_queue_id=None, service_id=None, instance_type=None, instance_seconds=None, cloud_provider=None, anyscale_credits=None, local_vars_configuration=None):  # noqa: E501
        """AggregatedInstanceUsageWithCostModel - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._date = None
        self._organization_id = None
        self._organization_name = None
        self._cloud_hosting_type = None
        self._cloud_id = None
        self._cloud_name = None
        self._project_id = None
        self._project_name = None
        self._user_id = None
        self._user_email = None
        self._cluster_id = None
        self._workspace_id = None
        self._job_id = None
        self._job_queue_id = None
        self._service_id = None
        self._instance_type = None
        self._instance_seconds = None
        self._cloud_provider = None
        self._anyscale_credits = None
        self.discriminator = None

        self.date = date
        self.organization_id = organization_id
        self.organization_name = organization_name
        self.cloud_hosting_type = cloud_hosting_type
        self.cloud_id = cloud_id
        self.cloud_name = cloud_name
        self.project_id = project_id
        self.project_name = project_name
        self.user_id = user_id
        self.user_email = user_email
        self.cluster_id = cluster_id
        if workspace_id is not None:
            self.workspace_id = workspace_id
        if job_id is not None:
            self.job_id = job_id
        if job_queue_id is not None:
            self.job_queue_id = job_queue_id
        if service_id is not None:
            self.service_id = service_id
        self.instance_type = instance_type
        self.instance_seconds = instance_seconds
        self.cloud_provider = cloud_provider
        self.anyscale_credits = anyscale_credits

    @property
    def date(self):
        """Gets the date of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The date of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: date
        """
        return self._date

    @date.setter
    def date(self, date):
        """Sets the date of this AggregatedInstanceUsageWithCostModel.


        :param date: The date of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: date
        """
        if self.local_vars_configuration.client_side_validation and date is None:  # noqa: E501
            raise ValueError("Invalid value for `date`, must not be `None`")  # noqa: E501

        self._date = date

    @property
    def organization_id(self):
        """Gets the organization_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The organization_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this AggregatedInstanceUsageWithCostModel.


        :param organization_id: The organization_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and organization_id is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_id`, must not be `None`")  # noqa: E501

        self._organization_id = organization_id

    @property
    def organization_name(self):
        """Gets the organization_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The organization_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._organization_name

    @organization_name.setter
    def organization_name(self, organization_name):
        """Sets the organization_name of this AggregatedInstanceUsageWithCostModel.


        :param organization_name: The organization_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and organization_name is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_name`, must not be `None`")  # noqa: E501

        self._organization_name = organization_name

    @property
    def cloud_hosting_type(self):
        """Gets the cloud_hosting_type of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The cloud_hosting_type of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: CloudHostingType
        """
        return self._cloud_hosting_type

    @cloud_hosting_type.setter
    def cloud_hosting_type(self, cloud_hosting_type):
        """Sets the cloud_hosting_type of this AggregatedInstanceUsageWithCostModel.


        :param cloud_hosting_type: The cloud_hosting_type of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: CloudHostingType
        """
        if self.local_vars_configuration.client_side_validation and cloud_hosting_type is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_hosting_type`, must not be `None`")  # noqa: E501

        self._cloud_hosting_type = cloud_hosting_type

    @property
    def cloud_id(self):
        """Gets the cloud_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The cloud_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this AggregatedInstanceUsageWithCostModel.


        :param cloud_id: The cloud_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def cloud_name(self):
        """Gets the cloud_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The cloud_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._cloud_name

    @cloud_name.setter
    def cloud_name(self, cloud_name):
        """Sets the cloud_name of this AggregatedInstanceUsageWithCostModel.


        :param cloud_name: The cloud_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_name is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_name`, must not be `None`")  # noqa: E501

        self._cloud_name = cloud_name

    @property
    def project_id(self):
        """Gets the project_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The project_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this AggregatedInstanceUsageWithCostModel.


        :param project_id: The project_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def project_name(self):
        """Gets the project_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The project_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._project_name

    @project_name.setter
    def project_name(self, project_name):
        """Sets the project_name of this AggregatedInstanceUsageWithCostModel.


        :param project_name: The project_name of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_name is None:  # noqa: E501
            raise ValueError("Invalid value for `project_name`, must not be `None`")  # noqa: E501

        self._project_name = project_name

    @property
    def user_id(self):
        """Gets the user_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The user_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id):
        """Sets the user_id of this AggregatedInstanceUsageWithCostModel.


        :param user_id: The user_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user_id is None:  # noqa: E501
            raise ValueError("Invalid value for `user_id`, must not be `None`")  # noqa: E501

        self._user_id = user_id

    @property
    def user_email(self):
        """Gets the user_email of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The user_email of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._user_email

    @user_email.setter
    def user_email(self, user_email):
        """Sets the user_email of this AggregatedInstanceUsageWithCostModel.


        :param user_email: The user_email of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user_email is None:  # noqa: E501
            raise ValueError("Invalid value for `user_email`, must not be `None`")  # noqa: E501

        self._user_email = user_email

    @property
    def cluster_id(self):
        """Gets the cluster_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The cluster_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this AggregatedInstanceUsageWithCostModel.


        :param cluster_id: The cluster_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_id`, must not be `None`")  # noqa: E501

        self._cluster_id = cluster_id

    @property
    def workspace_id(self):
        """Gets the workspace_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The workspace_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id):
        """Sets the workspace_id of this AggregatedInstanceUsageWithCostModel.


        :param workspace_id: The workspace_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """

        self._workspace_id = workspace_id

    @property
    def job_id(self):
        """Gets the job_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The job_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this AggregatedInstanceUsageWithCostModel.


        :param job_id: The job_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """

        self._job_id = job_id

    @property
    def job_queue_id(self):
        """Gets the job_queue_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The job_queue_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._job_queue_id

    @job_queue_id.setter
    def job_queue_id(self, job_queue_id):
        """Sets the job_queue_id of this AggregatedInstanceUsageWithCostModel.


        :param job_queue_id: The job_queue_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """

        self._job_queue_id = job_queue_id

    @property
    def service_id(self):
        """Gets the service_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The service_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._service_id

    @service_id.setter
    def service_id(self, service_id):
        """Sets the service_id of this AggregatedInstanceUsageWithCostModel.


        :param service_id: The service_id of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """

        self._service_id = service_id

    @property
    def instance_type(self):
        """Gets the instance_type of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The instance_type of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: str
        """
        return self._instance_type

    @instance_type.setter
    def instance_type(self, instance_type):
        """Sets the instance_type of this AggregatedInstanceUsageWithCostModel.


        :param instance_type: The instance_type of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and instance_type is None:  # noqa: E501
            raise ValueError("Invalid value for `instance_type`, must not be `None`")  # noqa: E501

        self._instance_type = instance_type

    @property
    def instance_seconds(self):
        """Gets the instance_seconds of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The instance_seconds of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: float
        """
        return self._instance_seconds

    @instance_seconds.setter
    def instance_seconds(self, instance_seconds):
        """Sets the instance_seconds of this AggregatedInstanceUsageWithCostModel.


        :param instance_seconds: The instance_seconds of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and instance_seconds is None:  # noqa: E501
            raise ValueError("Invalid value for `instance_seconds`, must not be `None`")  # noqa: E501

        self._instance_seconds = instance_seconds

    @property
    def cloud_provider(self):
        """Gets the cloud_provider of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The cloud_provider of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: CloudProviders
        """
        return self._cloud_provider

    @cloud_provider.setter
    def cloud_provider(self, cloud_provider):
        """Sets the cloud_provider of this AggregatedInstanceUsageWithCostModel.


        :param cloud_provider: The cloud_provider of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: CloudProviders
        """
        if self.local_vars_configuration.client_side_validation and cloud_provider is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_provider`, must not be `None`")  # noqa: E501

        self._cloud_provider = cloud_provider

    @property
    def anyscale_credits(self):
        """Gets the anyscale_credits of this AggregatedInstanceUsageWithCostModel.  # noqa: E501


        :return: The anyscale_credits of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :rtype: float
        """
        return self._anyscale_credits

    @anyscale_credits.setter
    def anyscale_credits(self, anyscale_credits):
        """Sets the anyscale_credits of this AggregatedInstanceUsageWithCostModel.


        :param anyscale_credits: The anyscale_credits of this AggregatedInstanceUsageWithCostModel.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and anyscale_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `anyscale_credits`, must not be `None`")  # noqa: E501

        self._anyscale_credits = anyscale_credits

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AggregatedInstanceUsageWithCostModel):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, AggregatedInstanceUsageWithCostModel):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CleanupLeakedGrafanaDashboardResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'deleted_dashboards': 'list[GrafanaDashboard]',
        'num_skipped': 'int',
        'num_active': 'int',
        'dry_run': 'bool'
    }

    attribute_map = {
        'deleted_dashboards': 'deleted_dashboards',
        'num_skipped': 'num_skipped',
        'num_active': 'num_active',
        'dry_run': 'dry_run'
    }

    def __init__(self, deleted_dashboards=None, num_skipped=None, num_active=None, dry_run=None, local_vars_configuration=None):  # noqa: E501
        """CleanupLeakedGrafanaDashboardResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._deleted_dashboards = None
        self._num_skipped = None
        self._num_active = None
        self._dry_run = None
        self.discriminator = None

        self.deleted_dashboards = deleted_dashboards
        self.num_skipped = num_skipped
        self.num_active = num_active
        self.dry_run = dry_run

    @property
    def deleted_dashboards(self):
        """Gets the deleted_dashboards of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501


        :return: The deleted_dashboards of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501
        :rtype: list[GrafanaDashboard]
        """
        return self._deleted_dashboards

    @deleted_dashboards.setter
    def deleted_dashboards(self, deleted_dashboards):
        """Sets the deleted_dashboards of this CleanupLeakedGrafanaDashboardResponse.


        :param deleted_dashboards: The deleted_dashboards of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501
        :type: list[GrafanaDashboard]
        """
        if self.local_vars_configuration.client_side_validation and deleted_dashboards is None:  # noqa: E501
            raise ValueError("Invalid value for `deleted_dashboards`, must not be `None`")  # noqa: E501

        self._deleted_dashboards = deleted_dashboards

    @property
    def num_skipped(self):
        """Gets the num_skipped of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501

        Number of dashboards skipped due to it not being an anyscale managed dashboard.  # noqa: E501

        :return: The num_skipped of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501
        :rtype: int
        """
        return self._num_skipped

    @num_skipped.setter
    def num_skipped(self, num_skipped):
        """Sets the num_skipped of this CleanupLeakedGrafanaDashboardResponse.

        Number of dashboards skipped due to it not being an anyscale managed dashboard.  # noqa: E501

        :param num_skipped: The num_skipped of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_skipped is None:  # noqa: E501
            raise ValueError("Invalid value for `num_skipped`, must not be `None`")  # noqa: E501

        self._num_skipped = num_skipped

    @property
    def num_active(self):
        """Gets the num_active of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501

        Number of dashboards skipped due to it being associated with an active entity.  # noqa: E501

        :return: The num_active of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501
        :rtype: int
        """
        return self._num_active

    @num_active.setter
    def num_active(self, num_active):
        """Sets the num_active of this CleanupLeakedGrafanaDashboardResponse.

        Number of dashboards skipped due to it being associated with an active entity.  # noqa: E501

        :param num_active: The num_active of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_active is None:  # noqa: E501
            raise ValueError("Invalid value for `num_active`, must not be `None`")  # noqa: E501

        self._num_active = num_active

    @property
    def dry_run(self):
        """Gets the dry_run of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501

        Whether this was a dry run or not. If so, the deleted dashboards list were not actually deleted.  # noqa: E501

        :return: The dry_run of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501
        :rtype: bool
        """
        return self._dry_run

    @dry_run.setter
    def dry_run(self, dry_run):
        """Sets the dry_run of this CleanupLeakedGrafanaDashboardResponse.

        Whether this was a dry run or not. If so, the deleted dashboards list were not actually deleted.  # noqa: E501

        :param dry_run: The dry_run of this CleanupLeakedGrafanaDashboardResponse.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and dry_run is None:  # noqa: E501
            raise ValueError("Invalid value for `dry_run`, must not be `None`")  # noqa: E501

        self._dry_run = dry_run

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CleanupLeakedGrafanaDashboardResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CleanupLeakedGrafanaDashboardResponse):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass, field
from typing import Dict, Union

import tzlocal

from anyscale._private.models import ModelBase, ModelEnum
from anyscale.job.models import JobConfig


UTC_TZ_NAME = "UTC"


@dataclass(frozen=True)
class ScheduleConfig(ModelBase):
    """Configuration options for a schedule."""

    __doc_py_example__ = """\
from anyscale.schedule.models import ScheduleConfig
from anyscale.job.models import JobConfig

config = ScheduleConfig(
    job_config=JobConfig(
        entrypoint="python main.py"
    ),
    timezone="local",
    cron_expression="0 0 * * *"
)
"""

    __doc_yaml_example__ = """\
timezone: local
cron_expression: 0 0 * * * *
job_config:
    name: my-schedule-job
    entrypoint: python main.py
    max_retries: 5
"""

    job_config: Union[Dict, JobConfig] = field(
        metadata={"docstring": "Configuration of the job."}
    )

    def _validate_job_config(self, job_config: Union[Dict, JobConfig]) -> JobConfig:
        if isinstance(job_config, dict):
            job_config = JobConfig.from_dict(job_config)

        if not isinstance(job_config, JobConfig):
            raise TypeError("'job_config' must be a JobConfig.")

        return job_config

    cron_expression: str = field(
        metadata={
            "docstring": "A cron expression to define the frequency at which to run this cron job, for example '0 0 * * *' is a cron expression that means 'run at midnight'. Visit crontab.guru to construct a precise cron_expression."
        }
    )

    def _validate_cron_expression(self, cron_expression: str):
        # validation of cron expression is primarily handled by the backend when
        # applying the schedule. simply check if it is a non-empty string here.
        if not isinstance(cron_expression, str):
            raise TypeError("'cron_expression' must be a string.")

        if not cron_expression:
            raise ValueError("'cron_expression' cannot be empty.")

    timezone: str = field(
        default=UTC_TZ_NAME,
        metadata={
            "docstring": "The timezone in which to interpret the cron_expression. Default is Universal time (UTC). You can specify 'local' to use the local timezone"
        },
    )

    def _validate_timezone(self, timezone: str) -> str:
        # validation of timezone is primarily handled by the backend when
        # applying the schedule. simply check if it is a non-empty string here.
        if not isinstance(timezone, str):
            raise TypeError("'timezone' must be a string.")
        if not timezone:
            raise ValueError("'timezone' cannot be empty.")
        elif timezone == "local":
            return tzlocal.get_localzone_name()
        else:
            return timezone


class ScheduleState(ModelEnum):
    """Current state of a schedule."""

    ENABLED = "ENABLED"
    DISABLED = "DISABLED"

    __docstrings__ = {
        ENABLED: "The schedule is enabled. Jobs will be started according to this jobs cron expression.",
        DISABLED: "The schedule is disabled. No jobs will be started until the schedule is reenabled.",
    }


@dataclass(frozen=True)
class ScheduleStatus(ModelBase):
    """Current status of a schedule."""

    __doc_py_example__ = """\
import anyscale
from anyscale.schedule.models import ScheduleStatus
status: ScheduleStatus = anyscale.schedule.status(name="my-schedule")
"""

    __doc_cli_example__ = """\
$ anyscale schedule status -n my-schedule
id: cronjob_dfhqufws6s3issltpjgrdzcgyc
name: my-schedule
state: ENABLED
"""

    config: ScheduleConfig = field(
        metadata={"docstring": "Configuration of the schedule."}
    )

    def _validate_config(self, config: ScheduleConfig):
        if not isinstance(config, ScheduleConfig):
            raise TypeError("'config' must be a ScheduleConfig.")

    id: str = field(
        metadata={
            "docstring": "Unique ID of the schedule (generated when the schedule is first submitted)."
        }
    )

    def _validate_id(self, id: str):  # noqa: A002
        if not isinstance(id, str):
            raise TypeError("'id' must be a string.")

    name: str = field(metadata={"docstring": "Name of the schedule."},)

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("'name' must be a string.")

    state: Union[str, ScheduleState] = field(
        metadata={"docstring": "Current state of the schedule."}
    )

    def _validate_state(self, state: Union[str, ScheduleState]) -> ScheduleState:
        return ScheduleState.validate(state)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class FinishFTJobRequestV2(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'ft_type': 'str',
        'model_config_version': 'str',
        'model_config': 'AviaryModelConfigV2'
    }

    attribute_map = {
        'ft_type': 'ft_type',
        'model_config_version': 'model_config_version',
        'model_config': 'model_config'
    }

    def __init__(self, ft_type=None, model_config_version=None, model_config=None, local_vars_configuration=None):  # noqa: E501
        """FinishFTJobRequestV2 - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._ft_type = None
        self._model_config_version = None
        self._model_config = None
        self.discriminator = None

        self.ft_type = ft_type
        self.model_config_version = model_config_version
        self.model_config = model_config

    @property
    def ft_type(self):
        """Gets the ft_type of this FinishFTJobRequestV2.  # noqa: E501


        :return: The ft_type of this FinishFTJobRequestV2.  # noqa: E501
        :rtype: str
        """
        return self._ft_type

    @ft_type.setter
    def ft_type(self, ft_type):
        """Sets the ft_type of this FinishFTJobRequestV2.


        :param ft_type: The ft_type of this FinishFTJobRequestV2.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ft_type is None:  # noqa: E501
            raise ValueError("Invalid value for `ft_type`, must not be `None`")  # noqa: E501
        allowed_values = ["lora", "full"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and ft_type not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `ft_type` ({0}), must be one of {1}"  # noqa: E501
                .format(ft_type, allowed_values)
            )

        self._ft_type = ft_type

    @property
    def model_config_version(self):
        """Gets the model_config_version of this FinishFTJobRequestV2.  # noqa: E501

        The version of LLMForge  # noqa: E501

        :return: The model_config_version of this FinishFTJobRequestV2.  # noqa: E501
        :rtype: str
        """
        return self._model_config_version

    @model_config_version.setter
    def model_config_version(self, model_config_version):
        """Sets the model_config_version of this FinishFTJobRequestV2.

        The version of LLMForge  # noqa: E501

        :param model_config_version: The model_config_version of this FinishFTJobRequestV2.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and model_config_version is None:  # noqa: E501
            raise ValueError("Invalid value for `model_config_version`, must not be `None`")  # noqa: E501

        self._model_config_version = model_config_version

    @property
    def model_config(self):
        """Gets the model_config of this FinishFTJobRequestV2.  # noqa: E501


        :return: The model_config of this FinishFTJobRequestV2.  # noqa: E501
        :rtype: AviaryModelConfigV2
        """
        return self._model_config

    @model_config.setter
    def model_config(self, model_config):
        """Sets the model_config of this FinishFTJobRequestV2.


        :param model_config: The model_config of this FinishFTJobRequestV2.  # noqa: E501
        :type: AviaryModelConfigV2
        """
        if self.local_vars_configuration.client_side_validation and model_config is None:  # noqa: E501
            raise ValueError("Invalid value for `model_config`, must not be `None`")  # noqa: E501

        self._model_config = model_config

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, FinishFTJobRequestV2):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, FinishFTJobRequestV2):
            return True

        return self.to_dict() != other.to_dict()



from datetime import datetime
from unittest.mock import Mock, patch

import anyscale
from anyscale._private.anyscale_client.anyscale_client import AnyscaleClient
from anyscale.client.openapi_client.models import Dataset as InternalDataset
from anyscale.llm.dataset import Dataset


def test_sdk():
    internal_dataset = InternalDataset(
        filename="test.jsonl",
        description="description",
        name="my_dataset",
        project_id="prj_123",
        id="dataset_123",
        creator_id="usr_123",
        created_at=datetime(2024, 1, 1),
        cloud_id="cld_123",
        num_versions=3,
        storage_uri="s3://bucket/path/to/test.jsonl",
        version=3,
    )

    def do_assertion(dataset: Dataset):
        assert dataset == Dataset(
            id="dataset_123",
            name="my_dataset",
            filename="test.jsonl",
            storage_uri="s3://bucket/path/to/test.jsonl",
            version=3,
            num_versions=3,
            created_at=datetime(2024, 1, 1),
            creator_id="usr_123",
            project_id="prj_123",
            cloud_id="cld_123",
            description="description",
        )

    with patch.multiple(
        AnyscaleClient,
        **{
            AnyscaleClient.__init__.__name__: Mock(return_value=None),
            AnyscaleClient.upload_dataset.__name__: Mock(return_value=internal_dataset),
            AnyscaleClient.get_dataset.__name__: Mock(return_value=internal_dataset),
            AnyscaleClient.download_dataset.__name__: Mock(return_value="hi".encode()),
        }
    ):
        dataset = anyscale.llm.dataset.upload("test.jsonl", "my_dataset")
        do_assertion(dataset)
        dataset = anyscale.llm.dataset.get("my_dataset")
        do_assertion(dataset)
        assert anyscale.llm.dataset.download("my_dataset").decode() == "hi"



from typing import List, Optional, Tuple

import click

from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.util import AnyscaleCommand
from anyscale.controllers.resource_quota_controller import ResourceQuotaController
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group("resource-quota", help="Anyscale resource quota commands.")
def resource_quota_cli() -> None:
    pass


@resource_quota_cli.command(
    name="create",
    help="Create a resource quota.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_CREATE_EXAMPLE,
)
@click.option(
    "-n", "--name", required=True, help="Name of the resource quota to create.",
)
@click.option(
    "--cloud",
    required=True,
    help="Name of the cloud that this resource quota applies to.",
)
@click.option(
    "--project",
    default=None,
    help="Name of the project that this resource quota applies to.",
)
@click.option(
    "--user-email",
    default=None,
    help="Email of the user that this resource quota applies to.",
)
@click.option(
    "--num-cpus",
    required=False,
    help="The quota limit for the number of CPUs.",
    type=int,
)
@click.option(
    "--num-instances",
    required=False,
    help="The quota limit for the number of instances.",
    type=int,
)
@click.option(
    "--num-gpus",
    required=False,
    help="The quota limit for the total number of GPUs.",
    type=int,
)
@click.option(
    "--num-accelerators",
    required=False,
    help="The quota limit for the number of accelerators. Example: --num-accelerators A100-80G 10",
    nargs=2,
    type=(str, int),
    multiple=True,
)
def create(  # noqa: PLR0913
    name: str,
    cloud: str,
    project: Optional[str],
    user_email: Optional[str],
    num_cpus: Optional[int],
    num_instances: Optional[int],
    num_gpus: Optional[int],
    num_accelerators: List[Tuple[str, int]],
) -> None:
    """Creates a resource quota.

    A name and cloud name must be provided.

    `$ anyscale resource-quota create -n my-resource-quota --cloud my-cloud --project my-project --user-email test@myorg.com --num-cpus 10 --num-instances 10 --num-gpus 10 --num-accelerators L4 5 --num-accelerators T4 10`
    """
    resource_quota_controller = ResourceQuotaController()
    resource_quota_controller.create(
        name=name,
        cloud=cloud,
        project=project,
        user_email=user_email,
        num_cpus=num_cpus,
        num_instances=num_instances,
        num_gpus=num_gpus,
        num_accelerators=dict(num_accelerators),
    )


@resource_quota_cli.command(
    name="list",
    help="List resource quotas.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_LIST_EXAMPLE,
)
@click.option(
    "-n", "--name", required=False, help="The name filter for the resource quotas.",
)
@click.option(
    "--cloud", required=False, help="The cloud filter for the resource quotas.",
)
@click.option(
    "--creator-id",
    required=False,
    help="The creator ID filter for the resource quotas.",
)
@click.option(
    "--is-enabled",
    required=False,
    default=None,
    help="The is_enabled filter for the resource quotas.",
    type=bool,
)
@click.option(
    "--max-items",
    required=False,
    default=20,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list_resource_quotas(
    name: Optional[str],
    cloud: Optional[str],
    creator_id: Optional[str],
    is_enabled: Optional[bool],
    max_items: int,
) -> None:
    """List resource quotas.

    `$ anyscale resource-quota list -n my-resource-quota --cloud my-cloud`
    """
    resource_quota_controller = ResourceQuotaController()
    resource_quota_controller.list_resource_quotas(
        name=name,
        cloud=cloud,
        creator_id=creator_id,
        is_enabled=is_enabled,
        max_items=max_items,
    )


@resource_quota_cli.command(
    name="delete",
    help="Delete a resource quota.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_DELETE_EXAMPLE,
)
@click.option(
    "--id", required=True, help="ID of the resource quota to delete.",
)
def delete(id: str) -> None:  # noqa: A002
    """Deletes a resource quota.

    An ID of resource quota must be provided.

    `$ anyscale resource-quota delete --id rsq_123`
    """
    resource_quota_controller = ResourceQuotaController()
    resource_quota_controller.delete(id)


@resource_quota_cli.command(
    name="enable",
    help="Enable a resource quota.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_ENABLE_EXAMPLE,
)
@click.option(
    "--id", required=True, help="ID of the resource quota to enable.",
)
def enable(id: str) -> None:  # noqa: A002
    """Enables a resource quota.

    An ID of resource quota must be provided.

    `$ anyscale resource-quota enable --id rsq_123`
    """
    resource_quota_controller = ResourceQuotaController()
    resource_quota_controller.set_status(id, is_enabled=True)


@resource_quota_cli.command(
    name="disable",
    help="Disable a resource quota.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_DISABLE_EXAMPLE,
)
@click.option(
    "--id", required=True, help="ID of the resource quota to disable.",
)
def disable(id: str) -> None:  # noqa: A002
    """Disables a resource quota.

    An ID of resource quota must be provided.

    `$ anyscale resource-quota disable --id rsq_123`
    """
    resource_quota_controller = ResourceQuotaController()
    resource_quota_controller.set_status(id, is_enabled=False)
